# OPEngine.pyw and OPEngine.py, version: 0.3.9, made for python 3.5, no non built-in python modules needed.
#
# Created by Alex Gardner
# Website: https://www.opsgames.co.uk
#
# This 3D game engine made in python can run games created by users.
# This should be imported as a module, and a game class be made in another file.
#
# Note to programmers: don't make your game engines in python, it's a bad idea - use java, C or C++ instead.
#    Tkinter's canvas isn't built for this sort of stuff, and not being able to render pixel by pixel makes
#    it difficult (near impossible) to render textures, shaders and lighting.
#

import tkinter as tk
import threading
import math
import time
import sys
import traceback

############## GLOBAL VARIABLES ##############

canvas_width = 800
canvas_height = 600
fov = 100.0

game = None
running = False
projection = None
master = None
w = None
bindings = []

class window:
    name = "OPEngine"
    version = "v0.3.9"

############## MATHS CLASSES ##############

class Vector3(object):
    def __init__(self,x,y,z):
        self.x = x
        self.y = y
        self.z = z
    def __str__(self):   return "("+str(self.x)+","+str(self.y)+","+str(self.z)+")"
    def length(self):    return math.sqrt((self.x*self.x) + (self.y*self.y) + (self.z*self.z))
    def dot(self,o):     return (self.x*o.x) + (self.y*o.y) + (self.z*o.z)
    def normalize(self):
        l = self.length()
        self.x/=l
        self.y/=l
        self.z/=l
        return self
    def rotate(self,angle,axis):
        sinHalfAngle = math.sin(math.radians(angle/2))
        cosHalfAngle = math.cos(math.radians(angle/2))
        x = axis.x * sinHalfAngle
        y = axis.y * sinHalfAngle
        z = axis.z * sinHalfAngle
        w = cosHalfAngle
        rot = Quaternion(x,y,z,w)
        conjugate = rot.conjugate()
        w = (rot*self)*conjugate
        self.x = w.x
        self.y = w.y
        self.z = w.z
        return
    def cross(self,o): return Vector3((self.y*o.z) - (self.z*o.y),(self.z*o.x) - (self.x*o.z),(self.x*o.y) - (self.y*o.x))
    def __add__(self,o):
        if(type(o) is Vector3): return Vector3(self.x+o.x,self.y+o.y,self.z+o.z)
        else:                   return Vector3(self.x+o,self.y+o,self.z+o)
    def __iadd__(self,o):
        if(type(o) is Vector3): self = Vector3(self.x+o.x,self.y+o.y,self.z+o.z)
        else:                   self = Vector3(self.x+o,self.y+o,self.z+o)
        return self
    def __sub__(self,o):
        if(type(o) is Vector3): return Vector3(self.x-o.x,self.y-o.y,self.z-o.z)
        else:                   return Vector3(self.x-o,self.y-o,self.z-o)
    def __isub__(self,o):
        if(type(o) is Vector3): self = Vector3(self.x-o.x,self.y-o.y,self.z-o.z)
        else:                   self = Vector3(self.x-o,self.y-o,self.z-o)
        return self
    def __mul__(self,o):
        if(type(o) is Vector3): return Vector3(self.x*o.x,self.y*o.y,self.z*o.z)
        else:                   return Vector3(self.x*o,self.y*o,self.z*o)
    def __imul__(self,o):
        if(type(o) is Vector3): self = Vector3(self.x*o.x,self.y*o.y,self.z*o.z)
        else:                   self = Vector3(self.x*o,self.y*o,self.z*o)
        return self
    def __div__(self,o):
        if(type(o) is Vector3): return Vector3(self.x/o.x,self.y/o.y,self.z/o.z)
        else:                   return Vector3(self.x/o,self.y/o,self.z/o)
    def __idiv__(self,o):
        if(type(o) is Vector3): self = Vector3(self.x/o.x,self.y/o.y,self.z/o.z)
        else:                   self = Vector3(self.x/o,self.y/o,self.z/o)
        return self
    def __truediv__(self,o):
        if(type(o) is Vector3): return Vector3(self.x/o.x,self.y/o.y,self.z/o.z)
        else:                   return Vector3(self.x/o,self.y/o,self.z/o)
    def __itruediv__(self,o):
        if(type(o) is Vector3): self = Vector3(self.x/o.x,self.y/o.y,self.z/o.z)
        else:                   self = Vector3(self.x/o,self.y/o,self.z/o)
        return self
class Vector4(object):
    def __init__(self,x,y,z,w):
        self.x = x
        self.y = y
        self.z = z
        self.w = w
    def __str__(self): return "("+str(self.x)+","+str(self.y)+","+str(self.z)+","+str(self.w)+")"
    def length(self):  return math.sqrt((self.x*self.x) + (self.y*self.y) + (self.z*self.z) + (self.w*self.w))
    def dot(self,o):   return (self.x*o.x) + (self.y*o.y) + (self.z*o.z) + (self.w*o.w)
    def cross(self,o):
        x = self.y*o.z - self.z*o.y
        y = self.z*o.x - self.x*o.z
        z = self.x*o.y - self.y*o.x
        return Vector4(x,y,z,0)
    def normalised(self):
        l = self.length()
        return Vector4(x/l,y/l,z/l,w/l)
    def rotate(self,angle,axis):
        sinAngle = math.sin(0-angle)
        cosAngle = math.cos(0-angle)
        return self.cross(axis*sinAngle)+(self*cosAngle)+(axis*self.dot(axis*(1-cosAngle)))
    def __add__(self,o):
        if(type(o) is Vector4): return Vector4(self.x+o.x,self.y+o.y,self.z+o.z,self.w+o.w)
        else:                   return Vector4(self.x+o,self.y+o,self.z+o,self.w+o)
    def __iadd__(self,o):
        if(type(o) is Vector4): self = Vector4(self.x+o.x,self.y+o.y,self.z+o.z,self.w+o.w)
        else:                   self = Vector4(self.x+o,self.y+o,self.z+o,self.w+o)
        return self
    def __sub__(self,o):
        if(type(o) is Vector4): return Vector4(self.x-o.x,self.y-o.y,self.z-o.z,self.w-o.w)
        else:                   return Vector4(self.x-o,self.y-o,self.z-o,self.w-o)
    def __isub__(self,o):
        if(type(o) is Vector4): self = Vector4(self.x-o.x,self.y-o.y,self.z-o.z,self.w-o.w)
        else:                   self = Vector4(self.x-o,self.y-o,self.z-o,self.w-o)
        return self
    def __mul__(self,o):
        if(type(o) is Vector4): return Vector4(self.x*o.x,self.y*o.y,self.z*o.z,self.w*o.w)
        else:                   return Vector4(self.x*o,self.y*o,self.z*o,self.w*o)
    def __imul__(self,o):
        if(type(o) is Vector4): self = Vector4(self.x*o.x,self.y*o.y,self.z*o.z,self.w*o.w)
        else:                   self = Vector4(self.x*o,self.y*o,self.z*o,self.w*o)
        return self
    def __div__(self,o):
        if(type(o) is Vector4): return Vector4(self.x/o.x,self.y/o.y,self.z/o.z,self.w/o.w)
        else:                   return Vector4(self.x/o,self.y/o,self.z/o,self.w/o)
    def __idiv__(self,o):
        if(type(o) is Vector4): self = Vector4(self.x/o.x,self.y/o.y,self.z/o.z,self.w/o.w)
        else:                   self = Vector4(self.x/o,self.y/o,self.z/o,self.w/o)
        return self
    def __truediv__(self,o):
        if(type(o) is Vector4): return Vector4(self.x/o.x,self.y/o.y,self.z/o.z,self.w/o.w)
        else:                   return Vector4(self.x/o,self.y/o,self.z/o,self.w/o)
    def __itruediv__(self,o):
        if(type(o) is Vector4): self = Vector4(self.x/o.x,self.y/o.y,self.z/o.z,self.w/o.w)
        else:                   self = Vector4(self.x/o,self.y/o,self.z/o,self.w/o)
        return self
class Matrix4(object):
    def __init__(self):
        self.m = [[0,0,0,0],[0,0,0,0],[0,0,0,0],[0,0,0,0]]
    def __str__(self):
        end = "[\n"
        for i in range(len(self.m)):
            for j in range(len(self.m[i])):
                end+=str(self.m[i][j])+", "
            end+="\n"
        end+="]"
        return end
    def initIdentity(self):
        self.m[0][0] = 1; self.m[0][1] = 0; self.m[0][2] = 0; self.m[0][3] = 0
        self.m[1][0] = 0; self.m[1][1] = 1; self.m[1][2] = 0; self.m[1][3] = 0
        self.m[2][0] = 0; self.m[2][1] = 0; self.m[2][2] = 1; self.m[2][3] = 0
        self.m[3][0] = 0; self.m[3][1] = 0; self.m[3][2] = 0; self.m[3][3] = 1
        return self
    def initScreenSpaceTransform(self,halfWidth,halfHeight):
        self.m[0][0] = halfWidth; self.m[0][1] = 0;            self.m[0][2] = 0; self.m[0][3] = halfWidth
        self.m[1][0] = 0;         self.m[1][1] = 0-halfHeight; self.m[1][2] = 0; self.m[1][3] = halfHeight
        self.m[2][0] = 0;         self.m[2][1] = 0;            self.m[2][2] = 1; self.m[2][3] = 0
        self.m[3][0] = 0;         self.m[3][1] = 0;            self.m[3][2] = 0; self.m[3][3] = 1
        return self
    def initPerspective(self, fov, aspectRatio,zNear,zFar):
        tanHalfFOV = math.tan(fov/2)
        zRange = zNear - zFar
        self.m[0][0] = 1.0/(tanHalfFOV*aspectRatio); self.m[0][1] = 0;              self.m[0][2] = 0;                     self.m[0][3] = 0
        self.m[1][0] = 0;                            self.m[1][1] = 1.0/tanHalfFOV; self.m[1][2] = 0;                     self.m[1][3] = 0
        self.m[2][0] = 0;                            self.m[2][1] = 0;              self.m[2][2] = (0-zNear-zFar)/zRange; self.m[2][3] = 2 * zFar * zNear / zRange
        self.m[3][0] = 0;                            self.m[3][1] = 0;              self.m[3][2] = 1;                     self.m[3][3] = 0
        return self
    def initTranslation(self,x,y,z):
        self.m[0][0] = 1; self.m[0][1] = 0; self.m[0][2] = 0; self.m[0][3] = x
        self.m[1][0] = 0; self.m[1][1] = 1; self.m[1][2] = 0; self.m[1][3] = y
        self.m[2][0] = 0; self.m[2][1] = 0; self.m[2][2] = 1; self.m[2][3] = z
        self.m[3][0] = 0; self.m[3][1] = 0; self.m[3][2] = 0; self.m[3][3] = 1
        return self
    def initRotation(self,x,y,z):
        rotX = Matrix4()
        rotY = Matrix4()
        rotZ = Matrix4()
        x = math.radians(x)
        y = math.radians(y)
        z = math.radians(z)
        
        rotZ.m[0][0] = math.cos(z); rotZ.m[0][1] = 0-math.sin(z); rotZ.m[0][2] = 0; rotZ.m[0][3] = 0
        rotZ.m[1][0] = math.sin(z); rotZ.m[1][1] = math.cos(z);   rotZ.m[1][2] = 0; rotZ.m[1][3] = 0
        rotZ.m[2][0] = 0;           rotZ.m[2][1] = 0;             rotZ.m[2][2] = 1; rotZ.m[2][3] = 0
        rotZ.m[3][0] = 0;           rotZ.m[3][1] = 0;             rotZ.m[3][2] = 0; rotZ.m[3][3] = 1
        
        rotX.m[0][0] = 1; rotX.m[0][1] = 0;           rotX.m[0][2] = 0;             rotX.m[0][3] = 0
        rotX.m[1][0] = 0; rotX.m[1][1] = math.cos(x); rotX.m[1][2] = 0-math.sin(x); rotX.m[1][3] = 0
        rotX.m[2][0] = 0; rotX.m[2][1] = math.sin(x); rotX.m[2][2] = math.cos(x);   rotX.m[2][3] = 0
        rotX.m[3][0] = 0; rotX.m[3][1] = 0;           rotX.m[3][2] = 0;             rotX.m[3][3] = 1
        
        rotY.m[0][0] = math.cos(y); rotY.m[0][1] = 0; rotY.m[0][2] = 0-math.sin(y); rotY.m[0][3] = 0
        rotY.m[1][0] = 0;           rotY.m[1][1] = 1; rotY.m[1][2] = 0;             rotY.m[1][3] = 0
        rotY.m[2][0] = math.sin(y); rotY.m[2][1] = 0; rotY.m[2][2] = math.cos(y);   rotY.m[2][3] = 0
        rotY.m[3][0] = 0;           rotY.m[3][1] = 0; rotY.m[3][2] = 0;             rotY.m[3][3] = 1
        self.m = (rotZ*(rotY*rotX)).m
        return self
    def initScale(self,x,y,z):
        self.m[0][0] = x; self.m[0][1] = 0; self.m[0][2] = 0; self.m[0][3] = 0
        self.m[1][0] = 0; self.m[1][1] = y; self.m[1][2] = 0; self.m[1][3] = 0
        self.m[2][0] = 0; self.m[2][1] = 0; self.m[2][2] = z; self.m[2][3] = 0
        self.m[3][0] = 0; self.m[3][1] = 0; self.m[3][2] = 0; self.m[3][3] = 1
        return self
    def transform(self,o):
        return Vector4(self.m[0][0]*o.x + self.m[0][1]*o.y + self.m[0][2]*o.z + self.m[0][3]*o.w,
                       self.m[1][0]*o.x + self.m[1][1]*o.y + self.m[1][2]*o.z + self.m[1][3]*o.w,
                       self.m[2][0]*o.x + self.m[2][1]*o.y + self.m[2][2]*o.z + self.m[2][3]*o.w,
                       self.m[3][0]*o.x + self.m[3][1]*o.y + self.m[3][2]*o.z + self.m[3][3]*o.w)
    def __mul__(self,o):
        result = Matrix4()
        for i in range(len(self.m)):
            for j in range(len(self.m[i])):
                result.m[i][j] = (self.m[i][0] * o.m[0][j] +
                                  self.m[i][1] * o.m[1][j] +
                                  self.m[i][2] * o.m[2][j] +
                                  self.m[i][3] * o.m[3][j])
        return result

class Quaternion(object):
    def __init__(self, x, y, z, w):
        self.x = x
        self.y = y
        self.z = z
        self.w = w
    def __str__(self): return "("+str(self.x)+","+str(self.y)+","+str(self.z)+","+str(self.w)+")"
    def length(self):   return math.sqrt((self.x*self.x) + (self.y*self.y) + (self.z*self.z) + (self.w*self.w))
    def normalize(self):
        l = self.length()
        self.x/=l
        self.y/=l
        self.z/=l
        self.w/=l
        return self
    def __mul__(self,o):
        if(type(o) is Quaternion):
            w = self.w*o.w - self.x*o.x - self.y*o.y - self.z*o.z
            x = self.x*o.w + self.w*o.x + self.y*o.z + self.z*o.y
            y = self.z*o.w + self.w*o.y + self.z*o.x + self.x*o.z
            z = self.z*o.w + self.w*o.z + self.x*o.y + self.y*o.x
            return Quaternion(x,y,z,w)
        if(type(o) is Vector3):
            w = (0-self.x)*o.x - self.y*o.y - self.z*o.z
            x = self.w*o.x     + self.y*o.z - self.z*o.y
            y = self.w*o.y     + self.z*o.x - self.x*o.z
            z = self.w*o.z     + self.x*o.y - self.y*o.x
            return Quaternion(x,y,z,w)
    def conjugate(self):
        return Quaternion(0-self.x,0-self.y,0-self.z,self.w)

############## ENGINE CLASSES ##############

class Vertex(object):
    def __init__(self,x):
        self.id=-1
        if(type(x) is Vector4):
            self.mpos=x
        else:
            self.mpos = Vector4(x[0],x[1],x[2],1)
    def __str__(self): return "("+str(self.mpos.x)+","+str(self.mpos.y)+","+str(self.mpos.z)+","+str(self.mpos.w)+")"
    def transform(self,trans):
        return Vertex(trans.transform(self.mpos))
    def perspectiveDivide(self):
        return Vertex(Vector4(self.mpos.x/self.mpos.w,self.mpos.y/self.mpos.w,self.mpos.z/self.mpos.w,self.mpos.w))

class GameObject(object):
    global projection
    def __init__(self, name, position, rotation, scale, colour):
        self.name = name
        self.position = position
        self.rotation = rotation
        self.scale = scale
        self.colour = colour
        self.elements = []
        self.translation = Matrix4().initTranslation(self.position.x,self.position.y,self.position.z)
        self.rotationM = Matrix4().initRotation(self.rotation.x, self.rotation.y, self.rotation.z)
        self.scaleM = Matrix4().initScale(self.scale.x,self.scale.y,self.scale.z)
        self.transform = projection*((self.translation*self.rotationM)*self.scaleM)
    def translate(self,movement):
        self.position += movement
        self.translation = Matrix4().initTranslation(self.position.x,self.position.y,self.position.z)
        self.transform = projection*((self.translation*self.rotationM)*self.scaleM)
    def rotate(self,rotation):
        self.rotation += rotation
        self.rotationM = Matrix4().initRotation(self.rotation.x, self.rotation.y, self.rotation.z)
        self.transform = projection*((self.translation*self.rotationM)*self.scaleM)
    def setPos(self,pos):
        self.position = pos
        self.translation = Matrix4().initTranslation(self.position.x,self.position.y,self.position.z)
        self.transform = projection*((self.translation*self.rotationM)*self.scaleM)
    def setRot(self,rot):
        self.rotation = rot
        self.rotationM = Matrix4().initRotation(self.rotation.x, self.rotation.y, self.rotation.z)
        self.transform = projection*((self.translation*self.rotationM)*self.scaleM)
    def setScale(self,scale):
        self.scale = scale
        self.scaleM = Matrix4().initScale(math.fabs(self.scale.x),math.fabs(self.scale.y),math.fabs(self.scale.z)) # prevents negative scale
        self.transform = projection*((self.translation*self.rotationM)*self.scaleM)
    def update(self):
        self.translation = Matrix4().initTranslation(self.position.x,self.position.y,self.position.z)
        self.rotationM = Matrix4().initRotation(self.rotation.x, self.rotation.y, self.rotation.z)
        self.scaleM = Matrix4().initScale(self.scale.x,self.scale.y,self.scale.z)
        self.transform = projection*((self.translation*self.rotationM)*self.scaleM)
        
class Cube(GameObject):
    def __init__(self, name, position, rotation, scale, colour, wireFrame=False,border=None):
        GameObject.__init__(self, name, position, rotation, scale, colour)
        self.type      = "Cube"
        self.wireFrame = wireFrame
        self.border    = border
        self.v1 = Vertex((-1,-1,1))
        self.v2 = Vertex((-1,1,1))
        self.v3 = Vertex((1,-1,1))
        self.v4 = Vertex((1,1,1))
        self.v5 = Vertex((-1,-1,-1))
        self.v6 = Vertex((-1,1,-1))
        self.v7 = Vertex((1,-1,-1))
        self.v8 = Vertex((1,1,-1))
    def render(self):
        last = self.elements
        self.elements = []
        if(self.wireFrame):
            self.elements.append(createSquare(self.v1,self.v2,self.v3,self.v4,transform = self.transform,wireFrame=True,colour = self.colour))
            self.elements.append(createSquare(self.v1,self.v2,self.v5,self.v6,transform = self.transform,wireFrame=True,colour = self.colour))
            self.elements.append(createSquare(self.v5,self.v6,self.v7,self.v8,transform = self.transform,wireFrame=True,colour = self.colour))
            self.elements.append(createSquare(self.v3,self.v4,self.v7,self.v8,transform = self.transform,wireFrame=True,colour = self.colour))
        else:
            screenSpaceTransform = Matrix4().initScreenSpaceTransform(canvas_width/2,canvas_height/2)
            V = [self.v1.transform(self.transform).transform(screenSpaceTransform).perspectiveDivide(),
            self.v2.transform(self.transform).transform(screenSpaceTransform).perspectiveDivide(), # TODO: move all transformations here,
            self.v3.transform(self.transform).transform(screenSpaceTransform).perspectiveDivide(), # so that we can compare the Z or W values
            self.v4.transform(self.transform).transform(screenSpaceTransform).perspectiveDivide(), # to determine the distance from camera
            self.v5.transform(self.transform).transform(screenSpaceTransform).perspectiveDivide(), # to determine which squares to draw first.
            self.v6.transform(self.transform).transform(screenSpaceTransform).perspectiveDivide(),
            self.v7.transform(self.transform).transform(screenSpaceTransform).perspectiveDivide(), 
            self.v8.transform(self.transform).transform(screenSpaceTransform).perspectiveDivide()]
            self.chooseVerts(V)
        for i in last:
            for j in i:
                w.delete(j)
        w.update_idletasks()
    def chooseVerts(self, v):
        norenders = []
        for i in range(8):
            v[i].id=i
        tris = [[0,1,2,3],[0,1,4,5],[4,5,6,7],[2,3,6,7],[0,2,4,6],[1,3,5,7]]
        xys = []
        for i in range(8):
            a = [i]
            a.append(int(v[i].mpos.x))
            a.append(int(v[i].mpos.y))
            a.append(v[i].mpos.z)
            xys.append(a)
        
        newpoints = sorted(xys, key=lambda a: a[3])
        norenders.append(newpoints[7][0])
        x=0
        current = []
        found = []
        while True:
            current.append(newpoints[x][0])
            for i in tris:
                if(found!=[]):
                    break
                for j in current:
                    if(i[0] in current and i[1] in current and i[2] in current and i[3] in current):
                        found = i
                        break
            x+=1
            if(x>=len(newpoints)):
               break
        del tris[tris.index(found)]
        self.crS(v,found,self.colour) # This draws the closest square in white
        for i in tris:
            t = True
            for j in norenders:
                if j in i:
                    t = False
                    break
            if(t):
                self.crS(v,i,self.colour)
    def getMaxMin(self,v,points1,points2):#unused
        p1 = []
        for i in points1:
            p1.append(v[i])
        p2 = []
        for i in points2:
            p2.append(v[2])
        xcoords1 = []
        ycoords1 = []
        for i in p1:
            xcoords1.append(i.mpos.x)
            ycoords1.append(i.mpos.y)
        p1xmax = max(xcoords1)
        p1xmin = min(xcoords1)
        p1ymax = max(ycoords1)
        p1ymin = min(ycoords1)
        xcoords1 = []
        ycoords1 = []
        for i in p2:
            xcoords1.append(i.mpos.x)
            ycoords1.append(i.mpos.y)
        p2xmax = max(xcoords1)
        p2xmin = min(xcoords1)
        p2ymax = max(ycoords1)
        p2ymin = min(ycoords1)
        if(p1xmax<p2xmax) or (p1ymax<p2ymax) or (p1xmin>p2xmin) or (p1ymin>p2ymin):
            return True
        else:
            return False
    def crS(self,V,j,b):
        self.elements.append(createSquare(V[j[0]],V[j[1]],V[j[2]],V[j[3]],done=True,colour = b,border = self.border))

############## ENGINE FUNCTIONS ##############

def rgb(red, green, blue): return "#%02x%02x%02x" % (red, green, blue)
def createTriangle(v1,v2,v3,Transform,wireFrame=False,colour=rgb(0,0,0),border=None):
    screenSpaceTransform = Matrix4().initScreenSpaceTransform(canvas_width/2,canvas_height/2)
    minV = v1.transform(Transform).transform(screenSpaceTransform).perspectiveDivide()
    midV = v2.transform(Transform).transform(screenSpaceTransform).perspectiveDivide()
    maxV = v3.transform(Transform).transform(screenSpaceTransform).perspectiveDivide()
    elements = []
    if(wireFrame):
        elements.append(w.create_line(minV.mpos.x,minV.mpos.y,midV.mpos.x,midV.mpos.y,fill=colour))
        elements.append(w.create_line(maxV.mpos.x,maxV.mpos.y,midV.mpos.x,midV.mpos.y,fill=colour))
        elements.append(w.create_line(maxV.mpos.x,maxV.mpos.y,minV.mpos.x,minV.mpos.y,fill=colour))
    else:
        elements.append(w.create_polygon(minV.mpos.x,minV.mpos.y,midV.mpos.x,midV.mpos.y,maxV.mpos.x,maxV.mpos.y,fill=colour,outline=border))
    return elements
def createSquare(v1,v2,v3,v4,done=False,transform = None,wireFrame=False,colour=rgb(0,0,0),border=None):
    if(done):
        aV = v1
        bV = v2
        cV = v3
        dV = v4
    else:
        screenSpaceTransform = Matrix4().initScreenSpaceTransform(canvas_width/2,canvas_height/2)
        aV = v1.transform(screenSpaceTransform).perspectiveDivide()
        bV = v2.transform(screenSpaceTransform).perspectiveDivide()
        cV = v3.transform(screenSpaceTransform).perspectiveDivide()
        dV = v4.transform(screenSpaceTransform).perspectiveDivide()
    elements = []
    if(wireFrame):
        elements.append(w.create_line(aV.mpos.x,aV.mpos.y,bV.mpos.x,bV.mpos.y,fill=colour))
        elements.append(w.create_line(bV.mpos.x,bV.mpos.y,dV.mpos.x,dV.mpos.y,fill=colour))
        elements.append(w.create_line(aV.mpos.x,aV.mpos.y,cV.mpos.x,cV.mpos.y,fill=colour))
        elements.append(w.create_line(cV.mpos.x,cV.mpos.y,dV.mpos.x,dV.mpos.y,fill=colour))
    else:
        elements.append(w.create_polygon(aV.mpos.x,aV.mpos.y,bV.mpos.x,bV.mpos.y,dV.mpos.x,dV.mpos.y,cV.mpos.x,cV.mpos.y,fill=colour,outline=border))
    return elements
def bindkey(key, action):
    global bindings
    bindings.append((key,action))

############## WINDOW CODE ##############

def shutdown():
    global running
    running = False
def config(event):
    global projection
    global canvas_width
    global canvas_height
    canvas_width = event.width
    canvas_height = event.height
    projection = Matrix4().initPerspective(math.radians(fov),canvas_width/canvas_height,0.1,1000)


############## MAIN ENGINE LOOP ##############

def runThread():
    global running
    global projection
    global game
    global bindings
    try:
        previousTime = time.time()
        projection = Matrix4().initPerspective(math.radians(fov),canvas_width/canvas_height,0.1,1000)
        running = True
        game._start()
        for i in bindings:
            w.bind(i[0],i[1])
        lastdeltas = []
        while True:
            currentTime = time.time()
            deltaTime = (currentTime - previousTime)
            while deltaTime==0:
                currentTime = time.time()
                deltaTime = (currentTime - previousTime)
            previousTime = currentTime
            if(len(lastdeltas)>=60):
                del lastdeltas[0]
            lastdeltas.append(1/deltaTime)
            average = sum(lastdeltas)/len(lastdeltas)
            master.title(window.name+" "+window.version+" - "+game.name+" "+game.version+" - FPS: "+str(int(average*10)/10.0))
            if(running == False):
                game._end()
                master.destroy()
                break
            game._update(deltaTime)
            
            game._render()
    except:
        print(sys.exc_info())
        traceback.print_tb(sys.exc_info()[2])
        running = False
        game._end()
        master.destroy()

############## BASE GAME ##############

class Game(object):
    def __init__(self,name="Untitled Game",version="v1",bgcolour=rgb(217,249,252)):
        self.objects = {}
        self.name = name
        self.version = version
        self.bgcolour = bgcolour
        self.uptime = 0.0
        self.deltaTime = 0.0
    def _start(self):
        w.config(bg=self.bgcolour)
        self.start()
    def _update(self,deltaTime):
        self.uptime += deltaTime
        self.update(deltaTime)
        self.deltaTime = deltaTime
    def _end(self):
        self.end()
    def _render(self):
        self.render()
    def start(self):
        pass #Override me!
    def update(self,deltaTime):
        pass #Override me!
    def end(self):
        pass #Override me!
    def render(self): #Override if game needs specific rendering instructions
        for i in self.objects.values():
            i.render()



############## ENGINE THREADS ##############
def runGame(myGame):
    global game
    global master
    global w
    game = myGame
    master = tk.Tk()
    master.title(window.name+ " "+window.version)
    master.protocol("WM_DELETE_WINDOW", shutdown)
    master.minsize(width=400,height=300)
    w = tk.Canvas(master, width=canvas_width, height=canvas_height,borderwidth=0,highlightthickness=0)
    w.pack(expand = tk.YES, fill = tk.BOTH)
    w.config(bg=rgb(217,249,252))
    w.bind("<Configure>",config)
    w.focus_set()
    th = threading.Thread(target=runThread)
    th.start()
    master.mainloop()
if __name__ == "__main__":
    print("Please run a game, this is only an engine module.")
    
