import tkinter as tk
import threading
import math
import time

canvas_width = 800
canvas_height = 600
fov = 100.0

running = False
projection = None
class Vector3(object):
    def __init__(self,x,y,z):
        self.x = x
        self.y = y
        self.z = z
    def __str__(self):   return "("+str(self.x)+","+str(self.y)+","+str(self.z)+")"
    def length(self):    return math.sqrt((self.x*self.x) + (self.y*self.y) + (self.z*self.z))
    def dot(self,o):     return (self.x*o.x) + (self.y*o.y) + (self.z*o.z)
    def normalize(self):
        l = self.length()
        self.x/=l
        self.y/=l
        self.z/=l
        return self
    def rotate(self,angle,axis):
        sinHalfAngle = math.sin(math.radians(angle/2))
        cosHalfAngle = math.cos(math.radians(angle/2))
        x = axis.x * sinHalfAngle
        y = axis.y * sinHalfAngle
        z = axis.z * sinHalfAngle
        w = cosHalfAngle
        rot = Quaternion(x,y,z,w)
        conjugate = rot.conjugate()
        w = (rot*self)*conjugate
        self.x = w.x
        self.y = w.y
        self.z = w.z
        return
        
    def cross(self,o): return Vector3((self.y*o.z) - (self.z*o.y),(self.z*o.x) - (self.x*o.z),(self.x*o.y) - (self.y*o.x))
    def __add__(self,o):
        if(type(o) is Vector3): return Vector3(self.x+o.x,self.y+o.y,self.z+o.z)
        else:                   return Vector3(self.x+o,self.y+o,self.z+o)
    def __iadd__(self,o):
        if(type(o) is Vector3): self = Vector3(self.x+o.x,self.y+o.y,self.z+o.z)
        else:                   self = Vector3(self.x+o,self.y+o,self.z+o)
        return self
    def __sub__(self,o):
        if(type(o) is Vector3): return Vector3(self.x-o.x,self.y-o.y,self.z-o.z)
        else:                   return Vector3(self.x-o,self.y-o,self.z-o)
    def __isub__(self,o):
        if(type(o) is Vector3): self = Vector3(self.x-o.x,self.y-o.y,self.z-o.z)
        else:                   self = Vector3(self.x-o,self.y-o,self.z-o)
        return self
    def __mul__(self,o):
        if(type(o) is Vector3): return Vector3(self.x*o.x,self.y*o.y,self.z*o.z)
        else:                   return Vector3(self.x*o,self.y*o,self.z*o)
    def __imul__(self,o):
        if(type(o) is Vector3): self = Vector3(self.x*o.x,self.y*o.y,self.z*o.z)
        else:                   self = Vector3(self.x*o,self.y*o,self.z*o)
        return self
    def __div__(self,o):
        if(type(o) is Vector3): return Vector3(self.x/o.x,self.y/o.y,self.z/o.z)
        else:                   return Vector3(self.x/o,self.y/o,self.z/o)
    def __idiv__(self,o):
        if(type(o) is Vector3): self = Vector3(self.x/o.x,self.y/o.y,self.z/o.z)
        else:                   self = Vector3(self.x/o,self.y/o,self.z/o)
        return self
    def __truediv__(self,o):
        if(type(o) is Vector3): return Vector3(self.x/o.x,self.y/o.y,self.z/o.z)
        else:                   return Vector3(self.x/o,self.y/o,self.z/o)
    def __itruediv__(self,o):
        if(type(o) is Vector3): self = Vector3(self.x/o.x,self.y/o.y,self.z/o.z)
        else:                   self = Vector3(self.x/o,self.y/o,self.z/o)
        return self
class Vector4(object):
    def __init__(self,x,y,z,w):
        self.x = x
        self.y = y
        self.z = z
        self.w = w
    def __str__(self): return "("+str(self.x)+","+str(self.y)+","+str(self.z)+","+str(self.w)+")"
    def length(self):  return math.sqrt((self.x*self.x) + (self.y*self.y) + (self.z*self.z) + (self.w*self.w))
    def dot(self,o):   return (self.x*o.x) + (self.y*o.y) + (self.z*o.z) + (self.w*o.w)
    def cross(self,o):
        x = self.y*o.z - self.z*o.y
        y = self.z*o.x - self.x*o.z
        z = self.x*o.y - self.y*o.x
        return Vector4(x,y,z,0)
    def normalised(self):
        l = self.length()
        return Vector4(x/l,y/l,z/l,w/l)
    def rotate(self,angle,axis):
        sinAngle = math.sin(0-angle)
        cosAngle = math.cos(0-angle)
        return self.cross(axis*sinAngle)+(self*cosAngle)+(axis*self.dot(axis*(1-cosAngle)))
    def __add__(self,o):
        if(type(o) is Vector4): return Vector4(self.x+o.x,self.y+o.y,self.z+o.z,self.w+o.w)
        else:                   return Vector4(self.x+o,self.y+o,self.z+o,self.w+o)
    def __iadd__(self,o):
        if(type(o) is Vector4): self = Vector4(self.x+o.x,self.y+o.y,self.z+o.z,self.w+o.w)
        else:                   self = Vector4(self.x+o,self.y+o,self.z+o,self.w+o)
        return self
    def __sub__(self,o):
        if(type(o) is Vector4): return Vector4(self.x-o.x,self.y-o.y,self.z-o.z,self.w-o.w)
        else:                   return Vector4(self.x-o,self.y-o,self.z-o,self.w-o)
    def __isub__(self,o):
        if(type(o) is Vector4): self = Vector4(self.x-o.x,self.y-o.y,self.z-o.z,self.w-o.w)
        else:                   self = Vector4(self.x-o,self.y-o,self.z-o,self.w-o)
        return self
    def __mul__(self,o):
        if(type(o) is Vector4): return Vector4(self.x*o.x,self.y*o.y,self.z*o.z,self.w*o.w)
        else:                   return Vector4(self.x*o,self.y*o,self.z*o,self.w*o)
    def __imul__(self,o):
        if(type(o) is Vector4): self = Vector4(self.x*o.x,self.y*o.y,self.z*o.z,self.w*o.w)
        else:                   self = Vector4(self.x*o,self.y*o,self.z*o,self.w*o)
        return self
    def __div__(self,o):
        if(type(o) is Vector4): return Vector4(self.x/o.x,self.y/o.y,self.z/o.z,self.w/o.w)
        else:                   return Vector4(self.x/o,self.y/o,self.z/o,self.w/o)
    def __idiv__(self,o):
        if(type(o) is Vector4): self = Vector4(self.x/o.x,self.y/o.y,self.z/o.z,self.w/o.w)
        else:                   self = Vector4(self.x/o,self.y/o,self.z/o,self.w/o)
        return self
    def __truediv__(self,o):
        if(type(o) is Vector4): return Vector4(self.x/o.x,self.y/o.y,self.z/o.z,self.w/o.w)
        else:                   return Vector4(self.x/o,self.y/o,self.z/o,self.w/o)
    def __itruediv__(self,o):
        if(type(o) is Vector4): self = Vector4(self.x/o.x,self.y/o.y,self.z/o.z,self.w/o.w)
        else:                   self = Vector4(self.x/o,self.y/o,self.z/o,self.w/o)
        return self
class Matrix4(object):
    def __init__(self):
        self.m = [[0,0,0,0],[0,0,0,0],[0,0,0,0],[0,0,0,0]]
    def __str__(self):
        end = "[\n"
        for i in range(len(self.m)):
            for j in range(len(self.m[i])):
                end+=str(self.m[i][j])+", "
            end+="\n"
        end+="]"
        return end
    def initIdentity(self):
        self.m[0][0] = 1; self.m[0][1] = 0; self.m[0][2] = 0; self.m[0][3] = 0
        self.m[1][0] = 0; self.m[1][1] = 1; self.m[1][2] = 0; self.m[1][3] = 0
        self.m[2][0] = 0; self.m[2][1] = 0; self.m[2][2] = 1; self.m[2][3] = 0
        self.m[3][0] = 0; self.m[3][1] = 0; self.m[3][2] = 0; self.m[3][3] = 1
        return self
    def initScreenSpaceTransform(self,halfWidth,halfHeight):
        self.m[0][0] = halfWidth; self.m[0][1] = 0;            self.m[0][2] = 0; self.m[0][3] = halfWidth
        self.m[1][0] = 0;         self.m[1][1] = 0-halfHeight; self.m[1][2] = 0; self.m[1][3] = halfHeight
        self.m[2][0] = 0;         self.m[2][1] = 0;            self.m[2][2] = 1; self.m[2][3] = 0
        self.m[3][0] = 0;         self.m[3][1] = 0;            self.m[3][2] = 0; self.m[3][3] = 1
        return self
    def initPerspective(self, fov, aspectRatio,zNear,zFar):
        tanHalfFOV = math.tan(fov/2)
        zRange = zNear - zFar
        self.m[0][0] = 1.0/(tanHalfFOV*aspectRatio); self.m[0][1] = 0;              self.m[0][2] = 0;                     self.m[0][3] = 0
        self.m[1][0] = 0;                            self.m[1][1] = 1.0/tanHalfFOV; self.m[1][2] = 0;                     self.m[1][3] = 0
        self.m[2][0] = 0;                            self.m[2][1] = 0;              self.m[2][2] = (0-zNear-zFar)/zRange; self.m[2][3] = 2 * zFar * zNear / zRange
        self.m[3][0] = 0;                            self.m[3][1] = 0;              self.m[3][2] = 1;                     self.m[3][3] = 0
        return self
    def initTranslation(self,x,y,z):
        self.m[0][0] = 1; self.m[0][1] = 0; self.m[0][2] = 0; self.m[0][3] = x
        self.m[1][0] = 0; self.m[1][1] = 1; self.m[1][2] = 0; self.m[1][3] = y
        self.m[2][0] = 0; self.m[2][1] = 0; self.m[2][2] = 1; self.m[2][3] = z
        self.m[3][0] = 0; self.m[3][1] = 0; self.m[3][2] = 0; self.m[3][3] = 1
        return self
    def initRotation(self,x,y,z):
        rotX = Matrix4()
        rotY = Matrix4()
        rotZ = Matrix4()
        x = math.radians(x)
        y = math.radians(y)
        z = math.radians(z)
        
        rotZ.m[0][0] = math.cos(z); rotZ.m[0][1] = 0-math.sin(z); rotZ.m[0][2] = 0; rotZ.m[0][3] = 0
        rotZ.m[1][0] = math.sin(z); rotZ.m[1][1] = math.cos(z);   rotZ.m[1][2] = 0; rotZ.m[1][3] = 0
        rotZ.m[2][0] = 0;           rotZ.m[2][1] = 0;             rotZ.m[2][2] = 1; rotZ.m[2][3] = 0
        rotZ.m[3][0] = 0;           rotZ.m[3][1] = 0;             rotZ.m[3][2] = 0; rotZ.m[3][3] = 1
        
        rotX.m[0][0] = 1; rotX.m[0][1] = 0;           rotX.m[0][2] = 0;             rotX.m[0][3] = 0
        rotX.m[1][0] = 0; rotX.m[1][1] = math.cos(x); rotX.m[1][2] = 0-math.sin(x); rotX.m[1][3] = 0
        rotX.m[2][0] = 0; rotX.m[2][1] = math.sin(x); rotX.m[2][2] = math.cos(x);   rotX.m[2][3] = 0
        rotX.m[3][0] = 0; rotX.m[3][1] = 0;           rotX.m[3][2] = 0;             rotX.m[3][3] = 1
        
        rotY.m[0][0] = math.cos(y); rotY.m[0][1] = 0; rotY.m[0][2] = 0-math.sin(y); rotY.m[0][3] = 0
        rotY.m[1][0] = 0;           rotY.m[1][1] = 1; rotY.m[1][2] = 0;             rotY.m[1][3] = 0
        rotY.m[2][0] = math.sin(y); rotY.m[2][1] = 0; rotY.m[2][2] = math.cos(y);   rotY.m[2][3] = 0
        rotY.m[3][0] = 0;           rotY.m[3][1] = 0; rotY.m[3][2] = 0;             rotY.m[3][3] = 1
        self.m = (rotZ*(rotY*rotX)).m
        return self
    def initScale(self,x,y,z):
        self.m[0][0] = x; self.m[0][1] = 0; self.m[0][2] = 0; self.m[0][3] = 0
        self.m[1][0] = 0; self.m[1][1] = y; self.m[1][2] = 0; self.m[1][3] = 0
        self.m[2][0] = 0; self.m[2][1] = 0; self.m[2][2] = z; self.m[2][3] = 0
        self.m[3][0] = 0; self.m[3][1] = 0; self.m[3][2] = 0; self.m[3][3] = 1
        return self
    def transform(self,o):
        return Vector4(self.m[0][0]*o.x + self.m[0][1]*o.y + self.m[0][2]*o.z + self.m[0][3]*o.w,
                       self.m[1][0]*o.x + self.m[1][1]*o.y + self.m[1][2]*o.z + self.m[1][3]*o.w,
                       self.m[2][0]*o.x + self.m[2][1]*o.y + self.m[2][2]*o.z + self.m[2][3]*o.w,
                       self.m[3][0]*o.x + self.m[3][1]*o.y + self.m[3][2]*o.z + self.m[3][3]*o.w)
    def __mul__(self,o):
        result = Matrix4()
        for i in range(len(self.m)):
            for j in range(len(self.m[i])):
                result.m[i][j] = (self.m[i][0] * o.m[0][j] +
                                  self.m[i][1] * o.m[1][j] +
                                  self.m[i][2] * o.m[2][j] +
                                  self.m[i][3] * o.m[3][j])
        return result

class Quaternion(object):
    def __init__(self, x, y, z, w):
        self.x = x
        self.y = y
        self.z = z
        self.w = w
    def __str__(self): return "("+str(self.x)+","+str(self.y)+","+str(self.z)+","+str(self.w)+")"
    def length(self):   return math.sqrt((self.x*self.x) + (self.y*self.y) + (self.z*self.z) + (self.w*self.w))
    def normalize(self):
        l = self.length()
        self.x/=l
        self.y/=l
        self.z/=l
        self.w/=l
        return self
    def __mul__(self,o):
        if(type(o) is Quaternion):
            w = self.w*o.w - self.x*o.x - self.y*o.y - self.z*o.z
            x = self.x*o.w + self.w*o.x + self.y*o.z + self.z*o.y
            y = self.z*o.w + self.w*o.y + self.z*o.x + self.x*o.z
            z = self.z*o.w + self.w*o.z + self.x*o.y + self.y*o.x
            return Quaternion(x,y,z,w)
        if(type(o) is Vector3):
            w = (0-self.x)*o.x - self.y*o.y - self.z*o.z
            x = self.w*o.x     + self.y*o.z - self.z*o.y
            y = self.w*o.y     + self.z*o.x - self.x*o.z
            z = self.w*o.z     + self.x*o.y - self.y*o.x
            return Quaternion(x,y,z,w)
    def conjugate(self):
        return Quaternion(0-self.x,0-self.y,0-self.z,self.w)
class Vertex(object):
    def __init__(self,x):
        if(type(x) is Vector4):
            self.mpos=x
        else:
            self.mpos = Vector4(x[0],x[1],x[2],1)
    def __str__(self): return "("+str(self.mpos.x)+","+str(self.mpos.y)+","+str(self.mpos.z)+","+str(self.mpos.w)+")"
    def transform(self,trans):
        return Vertex(trans.transform(self.mpos))
    def perspectiveDivide(self):
        return Vertex(Vector4(self.mpos.x/self.mpos.w,self.mpos.y/self.mpos.w,self.mpos.z/self.mpos.w,self.mpos.w))
    def triangleAreaDoubled(self, b, c):
        x1 = b.x - self.mpos.x
        y1 = b.y - self.mpos.y
        x2 = c.x - self.mpos.x
        y2 = c.y - self.mpos.y
        return (x1*y2 - x2*y1)
def rgb(red, green, blue): return "#%02x%02x%02x" % (red, green, blue)
def createTriangle(v1,v2,v3,Transform,wireFrame=False,colour=rgb(0,0,0)):
    screenSpaceTransform = Matrix4().initScreenSpaceTransform(canvas_width/2,canvas_height/2)
    minV = v1.transform(Transform).transform(screenSpaceTransform).perspectiveDivide()
    midV = v2.transform(Transform).transform(screenSpaceTransform).perspectiveDivide()
    maxV = v3.transform(Transform).transform(screenSpaceTransform).perspectiveDivide()
    elements = []
    if(wireFrame):
        elements.append(w.create_line(minV.mpos.x,minV.mpos.y,midV.mpos.x,midV.mpos.y,fill=colour))
        elements.append(w.create_line(maxV.mpos.x,maxV.mpos.y,midV.mpos.x,midV.mpos.y,fill=colour))
        elements.append(w.create_line(maxV.mpos.x,maxV.mpos.y,minV.mpos.x,minV.mpos.y,fill=colour))
    else:
        elements.append(w.create_polygon(minV.mpos.x,minV.mpos.y,midV.mpos.x,midV.mpos.y,maxV.mpos.x,maxV.mpos.y,fill=colour))
    return elements
def createSquare(v1,v2,v3,v4,Transform,wireFrame=False,colour=rgb(0,0,0)):
    screenSpaceTransform = Matrix4().initScreenSpaceTransform(canvas_width/2,canvas_height/2)
    aV = v1.transform(Transform).transform(screenSpaceTransform).perspectiveDivide()
    bV = v2.transform(Transform).transform(screenSpaceTransform).perspectiveDivide()
    cV = v3.transform(Transform).transform(screenSpaceTransform).perspectiveDivide()
    dV = v4.transform(Transform).transform(screenSpaceTransform).perspectiveDivide()
    elements = []
    if(wireFrame):
        elements.append(w.create_line(aV.mpos.x,aV.mpos.y,bV.mpos.x,bV.mpos.y,fill=colour))
        elements.append(w.create_line(bV.mpos.x,bV.mpos.y,dV.mpos.x,dV.mpos.y,fill=colour))
        elements.append(w.create_line(aV.mpos.x,aV.mpos.y,cV.mpos.x,cV.mpos.y,fill=colour))
        elements.append(w.create_line(cV.mpos.x,cV.mpos.y,dV.mpos.x,dV.mpos.y,fill=colour))
    else:
        elements.append(w.create_polygon(aV.mpos.x,aV.mpos.y,bV.mpos.x,bV.mpos.y,dV.mpos.x,dV.mpos.y,cV.mpos.x,cV.mpos.y,fill=colour))
    return elements
class GameObject(object):
    global projection
    def __init__(self, name, position, rotation, scale, colour):
        self.name = name
        self.projection = projection
        self.position = position
        self.rotation = rotation
        self.scale = scale
        self.colour = colour
        self.elements = []
        self.translation = Matrix4().initTranslation(self.position.x,self.position.y,self.position.z)
        self.rotationM = Matrix4().initRotation(self.rotation.x, self.rotation.y, self.rotation.z)
        self.scaleM = Matrix4().initScale(self.scale.x,self.scale.y,self.scale.z)
        self.transform = self.projection*((self.translation*self.rotationM)*self.scaleM)
    def translate(self,movement):
        self.position += movement
        self.translation = Matrix4().initTranslation(self.position.x,self.position.y,self.position.z)
        self.transform = self.projection*((self.translation*self.rotationM)*self.scaleM)
    def rotate(self,rotation):
        self.rotation += rotation
        self.rotationM = Matrix4().initRotation(self.rotation.x, self.rotation.y, self.rotation.z)
        self.transform = self.projection*((self.translation*self.rotationM)*self.scaleM)
    def setPos(self,pos):
        self.position = pos
        self.translation = Matrix4().initTranslation(self.position.x,self.position.y,self.position.z)
        self.transform = self.projection*((self.translation*self.rotationM)*self.scaleM)
    def setRot(self,rot):
        self.rotation = rot
        self.rotationM = Matrix4().initRotation(self.rotation.x, self.rotation.y, self.rotation.z)
        self.transform = self.projection*((self.translation*self.rotationM)*self.scaleM)
    def setScale(self,scale):
        self.scale = scale
        self.scaleM = Matrix4().initScale(self.scale.x,self.scale.y,self.scale.z)
        self.transform = self.projection*((self.translation*self.rotationM)*self.scaleM)
        
class Cube(GameObject):
    def __init__(self, name, position, rotation, scale, colour, wireFrame=False):
        GameObject.__init__(self, name, position, rotation, scale, colour)
        self.type="Cube"
        self.wireFrame=wireFrame
        self.v1 = Vertex((-1,-1,1))
        self.v2 = Vertex((-1,1,1))
        self.v3 = Vertex((1,-1,1))
        self.v4 = Vertex((1,1,1))
        self.v5 = Vertex((-1,-1,-1))
        self.v6 = Vertex((-1,1,-1))
        self.v7 = Vertex((1,-1,-1))
        self.v8 = Vertex((1,1,-1))
    def render(self):
        last = self.elements
        self.elements = []
        if(self.wireFrame):
            self.elements.append(createSquare(self.v1,self.v2,self.v3,self.v4,self.transform,wireFrame=True,colour = self.colour))
            self.elements.append(createSquare(self.v1,self.v2,self.v5,self.v6,self.transform,wireFrame=True,colour = self.colour))
            self.elements.append(createSquare(self.v5,self.v6,self.v7,self.v8,self.transform,wireFrame=True,colour = self.colour))
            self.elements.append(createSquare(self.v3,self.v4,self.v7,self.v8,self.transform,wireFrame=True,colour = self.colour))
        else:
            self.elements.append(createSquare(self.v1,self.v2,self.v3,self.v4,self.transform,colour = self.colour))
            self.elements.append(createSquare(self.v1,self.v2,self.v5,self.v6,self.transform,colour = self.colour))
            self.elements.append(createSquare(self.v5,self.v6,self.v7,self.v8,self.transform,colour = self.colour))
            self.elements.append(createSquare(self.v3,self.v4,self.v7,self.v8,self.transform,colour = self.colour))
            self.elements.append(createSquare(self.v1,self.v3,self.v5,self.v7,self.transform,colour = self.colour))
            self.elements.append(createSquare(self.v2,self.v4,self.v6,self.v8,self.transform,colour = self.colour))
        for i in last:
            for j in i:
                w.delete(j)
        w.update_idletasks()
master = tk.Tk()
master.title("OPEngine v0.3.1")
def shutdown():
    global running
    running = False
master.protocol("WM_DELETE_WINDOW", shutdown)

w = tk.Canvas(master, width=canvas_width, height=canvas_height)
w.pack(expand = tk.YES, fill = tk.BOTH)
w.config(bg=rgb(217,249,252))

def runThread():
    global running
    global projection
    projection = Matrix4().initPerspective(math.radians(fov),canvas_width/canvas_height,0.1,1000)
    previousTime = time.time()
    running = True
    game = Game()
    game.start()
    lastframe = []
    uptime = 0.0
    frames = 0
    lastsecond = time.time()
    while True:
        currentTime = time.time()
        deltaTime = (currentTime - previousTime)
        previousTime = currentTime
        uptime+=deltaTime
        frames+=1
        if(time.time() - lastsecond>=2):
            print(frames/2,"FPS")
            lastsecond=time.time()
            frames = 0
        if(running == False):
            master.destroy()
            break
    
        game.update(deltaTime,uptime)
        
        for i in game.objects.values():
            i.render()
        
        

### GAME CODE ###
class Game(object):
    def __init__(self):
        self.objects = {}
    def start(self):
        self.objects["cube1"] = Cube("cube",Vector3(0,0,4),Vector3(0,0,0),Vector3(0.9,0.9,0.9),rgb(200,200,200))
        self.objects["cube1box"] = Cube("cubebox",Vector3(0,0,4),Vector3(0,0,0),Vector3(1,1,1),rgb(0,0,0),wireFrame=True)
        self.objects["cube2"] = Cube("cube",Vector3(-2.5,2.5,4),Vector3(45,0,20),Vector3(1,1,1),rgb(20,240,140))
        self.objects["cube2box"] = Cube("cube",Vector3(-2.5,2.5,4),Vector3(45,0,20),Vector3(1,1,1),rgb(0,0,0),wireFrame=True)
        self.objects["cube3"] = Cube("cube",Vector3(0,-3,4),Vector3(20,0,20),Vector3(0.5,0.5,0.5),rgb(240,20,140))
        self.objects["cube3box"] = Cube("cube",Vector3(0,-3,4),Vector3(20,300,45),Vector3(0.5,0.5,0.5),rgb(255,255,255),wireFrame=True)
        self.objects["cube4box"] = Cube("cube", Vector3(2,0,4),Vector3(0,0,0),Vector3(1,1,1),rgb(0,255,255),wireFrame=True)
    def update(self,deltaTime,uptime):
        self.objects["cube1"].rotate(Vector3(1,-3,0)*deltaTime*10)
        self.objects["cube1box"].setRot(self.objects["cube1"].rotation)
        self.objects["cube4box"].setPos(Vector3(math.cos(uptime)*3,math.sin(uptime)*3,4))
        self.objects["cube3"].setScale(Vector3((2+math.cos(uptime))/2,0.5,0.5))
        self.objects["cube2"].setScale(Vector3(math.sin(uptime),math.sin(uptime),math.sin(uptime)))
        self.objects["cube3box"].transform = self.objects["cube3"].transform
        self.objects["cube2box"].transform = self.objects["cube2"].transform
    def end(self):
        pass
#Engine loops
th = threading.Thread(target=runThread)
th.start()
master.mainloop()
